/// \file standalone.c -- VS1053 SCI-controlled Standalone MP3 Player
/// Copyright (C) 2005-2011 VLSI Solution Oy Tampere Finland

/*
  See standalone.h for feature #defines.

  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

- VS1053:
  + higher reference
- IN VS1053b:
  + SCLK = GPIDATA8
  + XCS  = GPIDATA9
  + SI   = GPIDATA10
  + XDCS = GPIDATA11

 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "standalone.h"

extern s_int16 __y needCheckNextHeader, haveCachedHeader;
auto void play_frame(void);
void putch(register __a0 int ch);

#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))

#ifndef NOCHECKFILETYPE
  #define CHECKFILETYPE CheckFileType1053
#endif/*NOCHECKFILETYPE*/


s_int16 hcShift;

//#define CTRL3_UPDATE_VOL        (1<<15) /*not needed for vs1053*/
#define CTRL3_BY_NAME           (1<<8) /*locate by name*/
#define CTRL3_RECORD_ON         (1<<7)
#define CTRL3_AT_END            (1<<6)
#define CTRL3_NO_NUMFILES       (1<<5)
#define CTRL3_PAUSE_ON          (1<<4)
#define CTRL3_FILE_READY        (1<<3)

#define CTRL3_PLAY_MODE_MASK    (3<<1)
#define CTRL3_PAUSE_AFTER_PLAY  (3<<1)
#define CTRL3_PAUSE_BEFORE_PLAY (2<<1)
#define CTRL3_LOOP_SONG         (1<<1)
#define CTRL3_PLAY_NEXT         (0<<1)

#define CTRL3_RANDOM_PLAY       (1<<0) /* SCI_AICTRL3 */

#ifndef GPIO_IDATA
  #define GPIO_IDATA GPIO_DATA
#endif


#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */



typedef enum {
    spi_pause=0,
    spi_file=1,
    spi_seek=2,
    spi_waitdata=3,
    spi_data=4,
    spi_zeros=5,
    spi_record=6,
    spi_waitima=7
} SPISTATE;

SPISTATE spiState, oldSpiState;
u_int16 spiCnt;
#ifdef TIMEOUT
u_int16 timeOut;
#endif

struct SPIX { /* 0x1816 do not touch! ASM uses the structure */
    s_int32 size;
    u_int32 sector;
    u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
    __y struct FRAGMENT *fragment;
    u_int32 targetSector; /*goto 6,7*/
} spix;

#ifdef AB_REPEAT //Not yet... Needs UI
struct SPIX spia;
u_int16 decodeTimea;
u_int32 sectorB;
u_int16 abMode;
#endif

#if 1
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg);
#else
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg) {
    register __b1 u_int16 t;
    #if 1
        puts("Cmd(");
	puthex(cmd);
	puthex(arg>>16);
	puthex(arg);
	puts(")");
    #endif
    /* some MMC brands require the extra clock pulses     */
    SpiSendClocks();
    SpiSendClocks();
    USEX(GPIO_ODATA) &= ~SPI_xCS2;
  
    while (SpiSendReceiveMmc(0xff00,8) != 0xff) {
	/* Unexpected busy signal from MMC */
    }
    SpiSendReceiveMmc(cmd, 16); /* top half normally 0xff */
    SpiSendReceiveMmc(arg>>16, 16);
    SpiSendReceiveMmc(arg, 16);
    t = SpiSendReceiveMmc(0x9500, 8);    /* Valid CRC for init, then don't care */
  
    cmd = 100;               /* timeout value, try max. 100 times */
    while (cmd && (t & 0x80)) { /* R1 response not detected */
	t = SpiSendReceiveMmc(0xff00, 8);
	cmd--;               /* decrease the timeout value */
    }
    return t;
}
#endif

#if 1
auto void ReadDiskSector(register __a u_int32 sector);
#else
auto void ReadDiskSector(register __a u_int32 sector) {
    register s_int16 i;
    register u_int16 *buffer = stream_buffer;
  
    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector&0x7fffffffUL)<<hcShift);
    do {
        i = SpiSendReceiveMmc(0xff00,8);
    } while (i == 0xff);

    #if 0
      if (i != 0xfe) {
          SpiSendClocks();/*USEX(GPIO_ODATA) |= SPI_xCS2;*/
          return;
      }
    #endif
    
    for (i=0; i<512; i+=2) {
        *buffer++ = SpiSendReceiveMmc(0xffff,16);
    }
    SpiSendReceiveMmc(0xffff,16); /* discard crc */

    /* generate some extra SPI clock edges to finish up the command */
    SpiSendClocks();
    SpiSendClocks();
    return;
}
#endif

#if 0
struct Fat {
    byte BS_jmpBoot[3];         /**< 0 x86 Boot Jump Code */
    byte BS_OEMName[8];         /**< 3 Formatter's name, usually "MSWIN4.1" */
      
    word BPB_BytsPerSec;        /**< 11 Bytes per sector (512) */
    byte BPB_SecPerClus;        /**< 13 Sectors per Cluster (1,2,4,8,..,128) */
    word BPB_RsvdSecCnt;        /**< 14 Reserved sectors (1 (32 for FAT32)) */
    byte BPB_NumFATs;           /**< 16 Number of FATs (2) */
    word BPB_RootEntCnt;        /**< 17 FAT12/16 n of root dir entries */
    word BPB_TotSec16;          /**< 19 Old sector count (0 for FAT32) */
    byte BPB_Media;             /**< 21 Media Type (eg 0xF8) */
    word BPB_FATSz16;           /**< 22 Size of one FAT16 in sectors */
    word BPB_SecPerTrk;         /**< 24 Old CHS Sectors Per Track */
    word BPB_NumHeads;          /**< 26 Old CSH Number of Heads */
    u_32 BPB_HiddSec;           /**< 28 n of sectors before this volume */
    u_32 BPB_TotSec32;          /**< 32 New sector count (0 for FAT12/16) */
     
    /** FAT type specific extensions */
    union Extensions{   
  
        /** FAT12/16 specific extensions to Bios Parameter Block*/
        struct Fat16Specific {    
          byte BS_DrvNum;         /**< 36 DOS INT13 Drive Number (0x80=HD)*/
          byte BS_Reserved1;      /**< 37 For WINNT; Format to 0 */
          byte BS_BootSig;        /**< 38 0x29 if next 3 fields are present */
          byte BS_VolID[4];       /**< 39 Volume ID (usually format datetime) */
          byte BS_VolLab[11];     /**< 43 Volume Label */
          byte BS_FilSysType[8];  /**< 54 Decorative name of fs, eg "FAT16   "*/
        } _16;
  
        /** FAT32 specific extensions to Bios Parameter Block*/
        struct Fat32Specific {    
          u_32 BPB_FATSz32;       /**< 36 Size of one FAT32 in sectors */
          word BPB_ExtFlags;      /**< 40 Flags; active FAT number etc */
          word BPB_FSVer;         /**< 42 File System Version (0x0000) */
          u_32 BPB_RootClus;      /**< 44 Start cluster of Root Dir. (2) */
          word BPB_FSInfo;        /**< 48 Start sector of FSINFO in Resvd area */
          word BPB_BkBootSec;     /**< 50 Sector in Resvd area for BkBoot (6) */
          byte BPB_Reserved[12];  /**< 52 Reserved, Always 0. */
          byte BS_DrvNum;         /**< 64 DOS INT13 Drive Number (0x80=HD) */
          byte BS_Reserved1;      /**< 65 For WINNT, Format to 0 */
          byte BS_BootSig;        /**< 66 0x29 if next 3 fields are present */
          byte BS_VolID[4];       /**< 67 Volume ID (usually format datetime) */
          byte BS_VolLab[11];     /**< 71 Volume Label */
          byte BS_FilSysType[8];  /**< 82 Decorative name of fs, eg "FAT32   "*/
        } _32;
      } ext;
    } fat;
#endif

#if 1
auto u_int16 GetByte(register __c0 u_int16 n);
#else
auto u_int16 GetByte(register __c0 u_int16 n) {
    register u_int16 *p = stream_buffer[n>>1];
    if (n & 1) {
	n = *p;
    } else {
	n = *p >> 8;
    }
    return n & 0xff;
}
#endif
#if 1
auto u_int16 GetWord(register __c0 u_int16 n);
#else
auto u_int16 GetWord(register __c0 u_int16 n) {
    return GetByte(n) | (GetByte(n+1)<<8);
}
#endif
#if 1
auto u_int32 GetLong(register __c0 u_int16 n);
#else
auto u_int32 GetLong(register __c0 u_int16 n) {
    return GetWord(n) | ((u_int32)GetWord(n+2)<<16);
}
#endif

extern __y u_int16 fileName[6];
#ifdef FATINFO_IN_Y
__y struct FATINFO fatInfo;
#else
struct FATINFO fatInfo;
#endif


#if 1
auto u_int16 InitFileSystem(void);
#else
auto u_int16 InitFileSystem(void) {
    /* Load MBR -- the first sector on disk */
    ReadDiskSector(0);
  
    if (GetWord(510) != 0xaa55)
	return 0xaa55;

    /* Now leave MBR and load sector 0 of partition */
    /* check for boot sector instead of partition table */
    if (GetLong(0x36)!= (((long)'1'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F') &&
        GetLong(0x52)!= (((long)'3'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F')
        ) {
        fatInfo.currentSector = GetLong(0x1c6);
        ReadDiskSector(fatInfo.currentSector);
    }
    {
        u_int16 BPB_BlksPerSec;

        /* Determine FAT Type (16/32) */
        /* This should be done better, but it'll do for now. */
        fatInfo.BPB_RootEntCnt = GetWord(17);
        /* FAT32 does not have separate root entries. */
        /* TODO: should detect FAT32 by BPB_FATSz16==0 */
        fatInfo.IS_FAT_32 = fatInfo.BPB_RootEntCnt ? 0 : 1;
        fatInfo.FilSysType = GetWord(54+3); /* 0x3231 for FAT12 */

        /* OK, let's calculate */
        /* First, let's get rid of the idea that we have byte addresses
           in the file system. Nope, let's only deal in physical disk
           sectors of 512 bytes. First we convert the FAT byter per sector
           value to "512B disk sectors per fat sector" value. */

        BPB_BlksPerSec = GetWord(11) >> 9;

        /* Then we adjust the Sector per Cluster to mean physical disk
           sectors. in 99% of the cases it is already so because bytes
           per sector almost always is 512 in FAT. Maximum cluster size
           is 65536 bytes (128 disk sectors). */

        fatInfo.fatSectorsPerCluster = GetByte(13)/*BPB_SecPerClus*/ * BPB_BlksPerSec;
        fatInfo.fatStart = fatInfo.currentSector +
            (u_int32)GetWord(14)/*BPB_RsvdSecCnt*/ * BPB_BlksPerSec;
        fatInfo.rootStart = GetWord(22);/*BPB_FATSz16*/
        if (fatInfo.rootStart == 0) {
            if (!fatInfo.IS_FAT_32) 
                return 0x0b; /* should be FAT32; can not find root directory */
            fatInfo.rootStart = GetLong(36);/*BPB_FATSz32*/
        }
        /* BPB_NumFATs * BPB_BlksPerSec fits in a word (8+7 bits) */
        fatInfo.rootStart = fatInfo.rootStart * (GetByte(16)/*BPB_NumFATs*/ *
                                                 BPB_BlksPerSec)
            + fatInfo.fatStart;
        fatInfo.dataStart = fatInfo.rootStart + (fatInfo.BPB_RootEntCnt >> 4)
            - (fatInfo.fatSectorsPerCluster << 1); /*first cluster is cluster 2*/
    }
    return 0;
}
#endif

#if 0
/** FAT/VFAT directory record union */
union DirRecordUnion {
    /** Standard directory entry */
    struct Entry {
      byte Name[11]; 0
      byte Attr;     11
      byte NTRes;    12
      byte CrtTimeTenth; 13
      word CrtTime;  14
      word CrtDate;  16
      word LstAccDate; 18
      word FstClusHi; 20
      word WrtTime;   22
      word WrtDate;   24
      word FstClusLo; 26
      u_32 FileSize;  28
    } entry;
  
    /** Extended directory entry */
    struct LongEntry {
      byte Ord;      0
      word Name1[5]; 1 /**< characters 1-5 */
      byte Attr;     11
      byte Type;     12 /**< entry type, zero=long name component */
      byte Chksum;   13
      word Name2[6]; 14 /**< characters 6-11 */
      word FstClusLO; 26 /**< zero for long entry */
      word Name3[2]; 28 /**< characters 12-13 */
    } longentry;
  };
#endif


//#define MAX_FRAGMENTS 28
__y struct FRAGMENT fragments[MAX_FRAGMENTS];

#if 1
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
				       register __b u_int32 fatCluster);
#else
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
				       register __b u_int32 fatCluster) {
    register __d0 u_int16 offset;
  
    /* fatCluster means now current cluster, starting from 0, not 2. */
    /* note that startcluster=2 is already compensated in fatInitGlobals */
    #if 0
        frag->size = 0;
        frag->start = fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;
    #endif
    goto l55; /* read first sector */
  
    while (1) {
	u_int32 t;
	if (fatInfo.IS_FAT_32) {
	    t = GetLong(offset) & 0x0fffffffUL;
	} else {
            t = GetWord(offset);
            if (t >= 0xfff8U)
                t = 0x0fffffffUL;
        }
        frag->size += fatInfo.fatSectorsPerCluster;
        ++fatCluster;
        if (frag->size >= 32768U || t != fatCluster) {
            /* if not sequential cluster, start next fragment */
            fatCluster = t;
            if (frag == &fragments[MAX_FRAGMENTS-1] ||
                t >= 0x0ffffff8UL) {
                frag->start |= LAST_FRAGMENT;
                return frag+1;
            }
            frag++;
        l55:
            frag->size = 0;
            frag->start =
                fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;

        } else {
            offset += (fatInfo.IS_FAT_32 ? 4 : 2);
            offset &= 511;
            if (offset != 0)
                continue;
        }
read:
        /* If FAT page is not already in memory, load it. */
        if (fatInfo.IS_FAT_32) {
            offset = ((u_int16)fatCluster & 0x7F) << 2;
            ReadDiskSector((fatCluster >> 7) + fatInfo.fatStart);
        } else {
            offset = ((u_int16)fatCluster & 0xff) << 1;
            ReadDiskSector((fatCluster >> 8) + fatInfo.fatStart);
        }
    }
}
#endif

u_int16 gFileNum[2];

#ifdef __VSDSP__
auto s_int16 OpenFile(register __c0 u_int16 fileNum);
#else
auto s_int16 OpenFile(register __c0 u_int16 fileNum) {
    register __i1 __y struct FRAGMENT *curFragment = &fragments[0];
    register __i0 __y struct FRAGMENT * __y nextFragment;
    
    /* Start at the start of root directory. */
    
    if (fatInfo.IS_FAT_32) {
	nextFragment = FragmentList(curFragment, 2);
    } else {
	curFragment->start = fatInfo.rootStart | LAST_FRAGMENT;
	curFragment->size  = (fatInfo.BPB_RootEntCnt >> 4); /* /sectorsPerCluster?*/
	nextFragment = curFragment + 1;
    }
    gFileNum[0] = fileNum;
    gFileNum[1] = 0;
    return HandleDir(curFragment, nextFragment);
}
#endif

#ifdef CHECKFILETYPE
#if 0
#define MKID(a,b,c) ((a)|((b)<<8)|((u_int32)(c)<<16))
auto s_int16 CheckFileType(register __a u_int32 name) {
    name &= 0x00ffffffUL;
    if (name == MKID('W','A','V'))
	return -1;
    if (name == MKID('M','P','3'))
	return -1;
    if (name == MKID('M','I','D'))
	return -1;
    if (name == MKID('W','M','A'))
	return -1;
    if (name == MKID('A','S','F'))
	return -1;
    if (name == MKID('A','A','C'))
	return -1;
    if (name == MKID('M','P','4'))
	return -1;
    if (name == MKID('M','4','A'))
	return -1;
    if (name == MKID('3','G','P'))
	return -1;
    if (name == MKID('3','G','2'))
	return -1;
    if (name == MKID('O','G','G'))
	return -1;
    return 0;
}
#endif/*1*/
#endif/*CHECKFILETYPE*/

/*147 words*/
auto s_int16 HandleDir(register __i2 __y struct FRAGMENT *curFragment,
		       __y struct FRAGMENT *nextFragment) {
    register __c1 int i = 0;
    register u_int32 currentSector = curFragment->start /*& 0x7fffffffUL */;
  
    #ifdef ENABLE_WATCHDOG
        USEX(WDOG_RESET) = WDOG_RESET_VAL;
    #endif
#pragma msg 36 off
    goto first;
  
    while (1) {
	register __d1 u_int16 fn = GetByte(i+0); /* first char of filename */
  
#pragma msg 36 on
	/* We are now looking at FAT directory structure. */
	/* Is current file a regular file? */
	if (fn == 0)
	    goto eod; /*end of directory*/
	if (fn != 0xe5) { /* Not deleted */
            #ifdef WITH_SUBDIRECTORIES
	    /* Is it a subdirectory? */
	    if ((GetByte(i+11/*Attr*/) & 0x10) && fn != '.') {
		/* TODO: do not call if FAT12! */
		if (HandleDir(nextFragment,
			      FragmentList(nextFragment,
					   ((u_int32)GetWord(i+20)<<16)+GetWord(i+26))) < 0)
		    return -1;
		memset(stream_buffer, 0, 512/2);
		ReadDiskSector(currentSector);/*reread sector*/
	    } else /* it was a subdirectory */
            #endif
		/* Attributes: NO directory, NO volume id, NO system, NO hidden */
	    if ((GetByte(i+11/*Attr*/) & 0xde) == 0
            #ifdef CHECKFILETYPE
                    /*
                      check supported files from name suffix:
                      vs1053: wav mp3 mid wma asf mp4 aac m4a ogg fla
                    */
		&& CHECKFILETYPE(GetLong(i+8))
            #endif
		) {
		u_int32 cluster =
		    ((u_int32)GetWord(i+20) << 16) + GetWord(i+26);
		/* It is a regular file. */

		/* If open-by-name, check name and open file if match. */
#pragma msg 240 off
		if (USEX(SCI_AICTRL3) & CTRL3_BY_NAME) {
		    register int j, t = 0;
		    register u_int16 *p = stream_buffer + i/2;
		    register __y u_int16 *d = fileName;
		    p[5] &= 0xff00; /* clear the attribute byte */
		    for (j=0; j<6; j++) {
			t |= *d++ ^ *p++;
		    }
		    if (t == 0) {
			USEX(SCI_AICTRL0) = gFileNum[1];
			goto thisfile;
		    }
		}
#pragma msg 240 on


                gFileNum[1]++; /* total file count */
                if (gFileNum[0]-- == 0) {
                thisfile:
		    /* Put the 8.3-character filename to memory so the user
		       can read it. Two bytes per word, big-endian. */
		    {   /* 13 words */
			register int j;
			register u_int16 *p = stream_buffer + i/2;
			register __y u_int16 *d = fileName;
			for (j=0;j<6;j++) {
			    *d++ = *p++;
			}
		    }

                    /* ------------ FILE FOUND ------------- */
                    fatInfo.fileSize = GetLong(i+28); /*do first!*/
                    FragmentList(&fragments[0], cluster);
                    return -1; /* File found, All OK return */
                }
            } /* normal file */
        } /* 0xe5 / 0x00 deleted or unused entry */
        i = (i + 32) & 511;
        if (i == 0) {
            currentSector++;
            if (--(curFragment->size) == 0) {
                if ((s_int16)(curFragment->start>>16) < 0) {
		eod:
		    USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME;
                    return (s_int16)gFileNum[1];
                }
                curFragment++;
                currentSector = curFragment->start /*& 0x7fffffffUL*/;
            }
        first:
	    memset(stream_buffer, 0, 512/2);
            ReadDiskSector(currentSector);
        }
    }
}


#ifdef HDATCHECK
  u_int16 hdatCheck;
#endif
extern __y u_int16 hwSampleRate;

/* Can be done for all formats, because IdleHook is run as the main thread
   and not in interrupts. Whenever IdleHook is called, there either is
   not enough stream data anyway, or the decoder is outputting samples
   and there is not enough space in the audio buffer (thus not reading
   stream). */
u_int16 __x * __x stream_save;

#if defined(SCI_KEYS)
__y struct {
    u_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
} ui;
#endif

#if defined(SCI_KEYS)
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n);
s_int16 ApplAddrButRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
		       register __a0 s_int16 n);
#endif

s_int16 ApplAddrSci(register __i0 s_int16 **d, register __a1 s_int16 mode,
		    register __a0 s_int16 n);
s_int16 ApplAddrRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
		    register __a0 s_int16 n);



#if 1
auto u_int32 NextSector(void);
#else
auto u_int32 NextSector(void) {
    spix.sector++;
    if (spix.fragSize-1 == 0) {
	if ((s_int32)spix.sector < 0) /* was last fragment */
	    return -1;
	spix.fragment++;
	spix.sector  = spix.fragment->start /*& 0x7fffffffUL*/;
	spix.fragSize = spix.fragment->size; /* *fatInfo.sectorsPerCluster*/
    } else {
	spix.fragSize = spix.fragSize-1;
    }
    return 0;
}
#endif

#if 1
auto u_int16 swap(register __a0 u_int16 val);
#else
auto u_int16 swap(register __a0 u_int16 val) {
    register u_int32 v = val<<8;
    return (u_int16)(v>>16)|(u_int16)v;
}
#endif


#ifdef SHUFFLE_PLAY
u_int16 shuffle_offset = 0;
#endif

/*
  Called before HALT whenever we have free time.
 */
void UserHook(void) {

#ifdef AB_REPEAT
	if (abnppin) {
		if (abMode == 0) {
			abMode = 1;
			spia = spix;
			decodeTimea = USEX(SCI_DECODE_TIME);
			sectorb = 0;
		} else if (abMode == 1 && spix.sector != spia.sector) {
			abMode = 2;
			sectorb = spix.sector;
			spix = spia;
			USEX(SCI_DECODE_TIME) = decodeTimea;
		} else {
			abMode = 0;
			sectorb = 0;
		}
	}
#endif

	/* play pause/continue */
	if ((USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON)) {
		if (spiState != spi_pause) {
			oldSpiState = spiState;
			spiState = spi_pause;
  
			stream_save = stream_wr_pointer;
			stream_wr_pointer = stream_rd_pointer+1;
			/* Does not matter if it is one off..
			StreamDiff() returns t	he right value regardless. */
		}
	} else {
		if (spiState == spi_pause) {
			spiState = oldSpiState;
			stream_wr_pointer = stream_save;
		}
	}
    
	if ((USEX(SCI_AICTRL0) & 0x8000U)) {
		/* new song */
		USEX(SCI_AICTRL0) &= 0x7fff; /* ack command */
		/* clear pause and file ready bits automatically */
		USEX(SCI_AICTRL3) &= ~(CTRL3_PAUSE_ON|CTRL3_FILE_READY|CTRL3_AT_END);
	
		/* Discard any unread data, including CRC, before starting again */
		{
			register int i;
			for (i=0; i<256+2; i++)
			SpiSendReceiveMmc(-1/*0xffff*/,16);
		}
		SpiSendClocks(); //raise chip select
		/* Jump to new track immediately */
		MyReset();
	}

	/* SCI-controlled version, but with additional key support.
	Keys are:
       GPIO4 for Pause/Play(short press) and Toggle Shuffle(long press), 
       GPIO5 for V+(short) and Next(long), 
       GPIO6 for V-(short) and Previous(long).

		Long press of GPIO4 toggles shuffle play. When shuffle
	    is activated, a new song is immediately selected.
    */
	#ifdef SCI_KEYS
	if (ui.Strobe) {
		ui.Strobe = 0;
		#ifdef KEYS_IDLE_LOW
		/* pulled low when idle, pulled high when pressed */
		ui.Data = USEX(GPIO_IDATA) & (7<<4);
		#else
		/* pulled high when idle, pulled low when pressed */
		ui.Data = ~USEX(GPIO_IDATA) & (7<<4);
		#endif
		//putch(ui.Data);

		if (ui.Data == ui.Key) { /* current state same as previous */
			if (ui.KeyTime < 12) {
				ui.	KeyTime++;
			} else if (ui.Key) {
				//putch(ui.KeyTime);
				if (ui.Key == (1<<5)) {
					/* next file */
					USEX(SCI_AICTRL0) = (USEX(SCI_AICTRL0) + 1) | 0x8000U;
					//putch('>');
				} else if (ui.Key == (1<<6)) {
					/* Previous file */
					if (USEX(SCI_DECODE_TIME) < 5) {
						if (USEX(SCI_AICTRL0) == 0) {
							/* if at first file, go to last file */
							USEX(SCI_AICTRL0) = (USEX(SCI_AICTRL1) - 1)
								| 0x8000U;
							//putch('');
						} else {
							USEX(SCI_AICTRL0) = (USEX(SCI_AICTRL0) - 1)
								| 0x8000U; /* previous file */
							//putch('<');
						}
					} else {
						/* if played 5 seconds or more,
						start the same file from the beginning */
						USEX(SCI_AICTRL0) |= 0x8000U;
					}
				} else if (ui.Key == (1<<4)) {
					/* toggle random play */
					USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
					if (USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY) {
						/* when shuffle activated, selects new song
						immediately. */
						USEX(SCI_AICTRL0) = 0x8000U;
					}
				}
				ui.KeyTime = -32767; /*all long-presses are one-shot*/
			}
		} else {
			if (ui.Data == 0) {
				if (ui.KeyTime > 0 && ui.KeyTime < 12) {
					/* was pressed < 0.75 seconds */
					if (ui.Key == (1<<4)) { /*pause/play*/
						USEX(SCI_AICTRL3) ^= CTRL3_PAUSE_ON;
					} else if (ui.Key == (1<<5)) {
						if (USEX(SCI_VOL) > 0x0202U) /*volume up*/
							USEX(SCI_VOL) -= 0x0202;
					} else if (ui.Key == (1<<6)) {
						if (USEX(SCI_VOL) < 0xc0c0U) /*volume down*/
							USEX(SCI_VOL) += 0x0202;
					}
				}
			}
			ui.Key = ui.Data;
			ui.KeyTime = 0;
		}
	}
	#else
	//lcc does not work without the above!
	#endif/*SCI_KEYS*/
	
	if (spiState == spi_file) { /* Play file from the beginning. */
		/* extra byte will be read if the file size is odd */
		spix.size = (fatInfo.fileSize+1)/2; /* convert to words */
		spiState = spi_seek;
setfrag:
		spix.fragment = &fragments[0];
		spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
		spix.fragSize = fragments[0].size; /*todo: *fatInfo.sectorsPerCluster*/
	}

	if (spiState == spi_seek) {
		#if 0 /* support rudimentary goto. TODO: change this to ff/rew using AICTRL2! */
		if (spix.targetSector &&
			(USEX(SCI_STATUS) & (1<<SCIST_DO_NOT_JUMP)) == 0) {
			#if 0
			putch('S');
			putch('e');
			putch('e');
			putch('k');
			putch(' ');
			#endif
			/* can not jump backwards */
			if (spix.targetSector >= (spix.fragment->start & 0x7fffffff)) {
				s_int32 skip = spix.targetSector - spix.sector;
				/*jump one fragment at a time if necessary*/
				if (skip >= spix.fragSize)
				skip = spix.fragSize-1;
				spix.sector += skip;
				spix.fragSize -= (s_int16)skip;
				spix.size -= skip*512/2;
	
				if (spix.sector == spix.targetSector)
				spix.targetSector = 0;
			}
			needCheckNextHeader = 1;
		}
		#endif /*goto*/
	
		/* Send the SD/MMC read block command */
		MmcCommand(MMC_READ_SINGLE_BLOCK|0x40, (spix.sector & 0x7fffffffUL) << hcShift);
		spiState = spi_waitdata;
		spiCnt = 0;
#ifdef AB_REPEAT
		if (sectorb && sectorb == spix.sector) {
			spix = spia;
			USEX(SCI_DECODE_TIME) = decodeTimea;
		}
#endif
		#ifdef HDATCHECK
		if (++hdatCheck == HDATCHECK && USEX(SCI_HDAT1) == 0) {
			/* if file not detected at HDATCHECK/2 kB, skip the rest */
			goto end_of_file;
		}
		#endif
	}
	
	/*TODO: handle spi_waitdata and spi_data
      in timer interrupt to increase throughput for FLAC */
	/*
      We have sent the read block command and now check if the data
      is available to be read out.
     */
	if (spiState == spi_waitdata) {
		register u_int16 res = SpiSendReceiveMmc(0xff00,8);
		if (res == 0xfe) { /* 0xfe is the data start token */
			spiState = spi_data; /* go to fetch mode */
			spiCnt = 0;
			#ifdef TIMEOUT
			timeOut = 0;
			#endif
			NextSector();
		} else if (res == 0xff) { /* card is still fetching block */
			if ((s_int16)++spiCnt < 0) {
				MyReset(); /* timeout, reset immediately */
			}
		} else {
			/* Otherwise we have Read Error, and data is not transferred */
			goto end_of_file;
		}
	} else if (spiState == spi_data) { /* data transfer state */
		/* Loop until the whole file is read, or stream buffer is full,
		or the next sector needs to be read. */
		while (spix.size > 0) {
			register int i, j = 32;
			register u_int16 *wr;
			#if defined(USE_FLAC)
			if (USEX(SCI_HDAT1) == (('f'<<8) | 'L')) {
				register s_int16 t = stream_wr_pointer - stream_rd_pointer;
				if (t < 0)
					t += 0x1800;
				if (t > 0x1800/*STREAM_BUFFER_SZ*/-40)
					break; /* stream buffer full, leave loop */
				wr = (void *)stream_wr_pointer;
				if (j > spix.size)
					j = (int)spix.size;
				for (i=0;i<j;i++) {
					*wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
					if (wr >= &stream_buffer[0x1800/*STREAM_BUFFER_SZ*/])
						wr = stream_buffer;
				}
			} else
			#endif
			{
				if (StreamDiff() >= STREAM_BUFFER_SZ-40)
					break; /* stream buffer full, leave loop */
				wr = (void *)stream_wr_pointer;
				if (j > spix.size)
					j = (int)spix.size;
				for (i=0;i<j;i++) {
					*wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
					if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
						wr = stream_buffer;
				}
			}
			stream_wr_pointer = (void *)wr;
			spix.size -= j;//32;
			spiCnt += 32;
			if (spiCnt >= 256) {
				/* 256 words read, get next block in file */
				spiState = spi_seek;
				SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
				break;
			}
		#ifdef TIMEOUT
			timeOut = 0;
		#endif
	    /* watchdog clearing here doesn't work with midi */
		}
  
#ifdef TIMEOUT
		if ((s_int16)++timeOut == -1) {
			spix.size = 0;
		}
#endif
    
#ifdef ENABLE_WATCHDOG
		USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
		if (spix.size <= 0) {
end_of_file:
			if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG){
				/* restart the same file immediately */
				spiState = spi_file;
			} else {
				/*NEW: pause after play */
				if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
					CTRL3_PAUSE_AFTER_PLAY) {
					USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON | CTRL3_AT_END;
				} else {
					USEX(SCI_AICTRL0) += 1; /* next file */
				}
				spiState = spi_zeros;
			}
		}
		#ifdef NO_WMA /* reject WMA files */
		if (USEX(SCI_HDAT1) == (('W'<<8)|'M')) {
			USEX(SCI_AICTRL0) += 1; /* next file */
			goto goreset;//MyReset();
		}
		#endif
		#ifdef NO_AAC /* reject AAC files */
		{
			register u_int16 t = USEX(SCI_HDAT1);
			//0x4154(AT) 0x4144(AD) 0x4d34(M4)
			if (t == 0x4154 || t == 0x4144 || t == 0x4d34) {
				USEX(SCI_AICTRL0) += 1; /* next file */
				goto goreset;//MyReset();
			}
		}
		#endif
	} else if (spiState == spi_zeros) { /* send zeros after the file */
		/* Workaround for WMA resync after stream ended! */
		parametric_x.resync = 0;
	
		/*TODO: different for flac */
		if (StreamDiff() < STREAM_BUFFER_SZ-40) {
			register int i, z = parametric_x.endFillByte;
			register u_int16 *wr = (void *)stream_wr_pointer;
			for (i=0;i<32;i++) {
				*wr = z;
				if (wr >= &stream_buffer[STREAM_BUFFER_SZ-1]) {
					wr = stream_buffer;
				} else {
					wr++;
				}
				++spiCnt;
				stream_wr_pointer = (void *)wr;
			}
			/* watchdog clearing here doesn't work with midi */
			#ifdef TIMEOUT
			timeOut = 0;
			#endif
		}
		#ifdef TIMEOUT
		if ((s_int16)++timeOut < 0) {
			//putch(USEX(SCI_HDAT1)>>8);
			//putch(USEX(SCI_HDAT1));
			goto goreset;
		}
		#endif
       	 	
		if (USEX(SCI_HDAT1) == 0 ||
		#if 1
			USEX(SCI_HDAT1) == 0x574d ||
		#endif
		#ifdef USE_FLAC
			USEX(SCI_HDAT1) == (('f'<<8) | 'L') ||
		#endif
			spiCnt >= 256+2*2048) { /* was left to max 256 by spi_data */
			/* 2048 words instead of bytes */
		goreset:
			/* Discard any unread data, including CRC, before starting again */
			{
				register int i;
				for (i=0; i<256+2; i++)
					SpiSendReceiveMmc(-1/*0xffff*/,16);
			}
			SpiSendClocks();

			MyReset();
		}
		#ifdef ENABLE_WATCHDOG
		goto wdogack;//USEX(WDOG_RESET) = WDOG_RESET_VAL;
		#endif
	} else {
		#ifdef ENABLE_WATCHDOG
  wdogack:
		USEX(WDOG_RESET) = WDOG_RESET_VAL;
    #endif
	}
	#ifdef SCI_KEYS
	if ((ui.Timer & 0x18) == 0) {
		/* DREQ indicates shuffle/random play status
		by blinking. */
		USEX(SER_DREQ) = USEX(SCI_AICTRL3);  /* lowest bit is random play*/
	} else
	#endif/*SCI_KEYS*/
	{
		USEX(SER_DREQ) = 0;
	}
}



auto int PatchFrame(void);
void AacSetRate(register __c1 u_int16 rate);/*Patches some AAC problems*/

/*
  TODO: use a timer interrupt for a watchdog timer.
 */
auto void MyMain(register s_int16 __a1 fullReset) {

#ifdef AB_REPEAT
	abMode = 0;
#endif

	/* If started without sensible defaults, set some. */
	if (USEX(SCI_CLOCKF) == 0) {
		USEX(SCI_CLOCKF) = 0x8000;
		USEX(SCI_VOL) = 0x2020;
	}

  /* Make sure we have the right upper ROM visible. */
  USEX(VS1053_IROM4) &= ~(IROM4_ENABLE);
  /* clear mp4ASC, seems to be needed after FLAC in vs1063, clear here also.*/
  memset((void *)0x1f00, 0, 64); /*Note: different address for vs1053 and 63*/
  USEX(SCI_HDAT1) = 0; /* clear so FLAC is not indicated. */

  #if defined(ENABLE_WATCHDOG)
    USEX(WDOG_CONFIG) = 1000;
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
  #endif

  /* no files yet */
  USEX(SCI_AICTRL1) = 0;
  memset(&spix, 0, sizeof(spix));

    /* For MMC/SD handling the SCI_NEWMODE must be set, and also
       the relevant GPIO pins configured correctly. */
//#define SD_POWER_ENABLE (1<<7)
#define SD_POWER_ENABLE 0
//#define LINE_IN_ENABLE (1<<6)
    USEX(SCI_MODE) = SciModeOr() | (USEX(SCI_MODE) & ~(1<<SCIMB_OUT_OF_WAV));
    USEX(GPIO_ODATA) = SPI_xCS|SPI_CLK|SPI_xCS2|SD_POWER_ENABLE;
    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2|SD_POWER_ENABLE
    #if defined(ENABLE_I2S)
	/* Enable I2S pins */
	| 0xf0;
      USEX(I2S_CONTROL) = 0x0c
    #endif
	;

    /* InitHardware() sets default UART speed to 9600bps,
       configures clock/PLL according to SCI_CLOCKF,
       initializes stream and audio buffer pointers,
       sets default samplerate (8000Hz),
       clears HDAT0,HDAT1,DECODE_TIME,AUDATA,
       and sets the cosmetic chip ID in parametric_x.chipID .
    */
    InitHardware(); /* no longer clears memory in vs1053 so preserves consts */

    /* Enable UART RX interrupt to be able to jump to emulator.
       Allows reflashing without MMC/SD inserted. */
    USEX(INT_ENABLE) = (1<<INT_EN_RX);

    InitLayer3(); /* clears mp3 synthesis buffers etc. */
    InitCommon(); /* sets mpeg frame pointers stuff */
    InitDecode(); /* sets synthesis phase variable */

    /* If not powerup reset, re-enable analog immediately. */
    if (!fullReset) {
	USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2);
    }
    #if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
      else {
	  /*analog power off and analog driver powerdown */
	  USEX(SCI_STATUS) |= (1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2);
      }
    #endif

#if defined(ENABLE_HIGHREF) && !defined(SCI_KEYS)
    /* Keeps settings 1.65V reference. */
    applAddr = ApplAddrRef;
#endif

#if defined(SCI_KEYS)
#if defined(ENABLE_HIGHREF)
    /* Calculates ui.Timer and ui.Strobe, keeps settings 1.65V reference.*/
    applAddr = ApplAddrButRef;
#else
    /* Calculates ui.Timer and ui.Strobe. */
    applAddr = ApplAddr;
#endif
#endif

    /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
    //USEX(INT_GLOB_ENA) = 0;

    /* wake up MMC! -- If this takes too long, watchdog will bite. */

    /* Initialize MMC/SD card, required for both player and recorder. */

    //while (1)
    {
	   register s_int16 i, cmd;

	// Start the MMC/SD card
    tryagain: 

#if defined(SD_POWER_ENABLE) && SD_POWER_ENABLE
	/* If required, power down MMC/SD to make it leave SD mode.
	   However, it should not be powered down at each tryagain,
	   because sometimes multiple passes are required to bring
	   the card into operating mode. */
	// MMC reset using SD_POWER_ENABLE
	USEX(GPIO_DDR) |= SD_POWER_ENABLE;
	USEX(SER_DREQ) = 0;
	{
	  u_int16 gpio0_save = USEX(GPIO_ODATA);
	  USEX(GPIO_ODATA) = 1; //all 0 except eecs
	  for (i=0; i<32000; i++) {
	    USEX(GPIO_ODATA) = 1; //all 0 except eecs
	  }
	  USEX(GPIO_ODATA) = gpio0_save | SD_POWER_ENABLE;
	}
#endif

	/* Send clocks to MMC/SD with chip select high
	   -- is required by most cards. */
	for (i=0; i<512; i++) {
	    SpiSendClocks();
	}

	/* MMC Init, command GO_IDLE_STATE should return 0x01 if all is ok. */
	i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
	if (i != 1) {
		goto tryagain;//continue; /* No valid idle response */
	}
	cmd = MMC_SEND_OP_COND|0x40;
	if (MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40, 0x00000122/*2.7-3.6V*/)
	    == 1) {
	    /* MMC answers: 0x05 SD answers: 0x01 */
	    /* Read the whole R7 response! */
	    SpiSendReceiveMmc(-1, 32);
	    cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
	}

        /* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
    //i = 0; /* i is 1 when entered .. but does not make the code shorter*/
    while (1) {
		register int c;
		if (cmd == (0x40|41)) {
			MmcCommand(0x40|55/*CMD55*/,0);
			c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0x40000000UL);
		} else {
			c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0);
		}
		if (c == 0)
			break;
		if (++i >= 12000/*< 0*/ || c != 1) {
			goto tryagain; /* Not able to power up mmc */
		}
	}

	hcShift = 9;
	/* PRETEC's 1G MMC does not seemd to handle CMD58 here! */
	/* Support HC for SD, but not for MMC! */
	if (cmd == (0x40|41) &&
		MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0) == 0) {
		if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
			/* OCR[30]:CCS - card capacity select */
			/* HighCapacity, change shift values */
			hcShift = 0;
	    }
	    SpiSendReceiveMmc(-1, 16); /* read the rest of the response */
	}
        
        #if 0
          /* Set Block Size of 512 bytes -- default for at least HC */
          /* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
          /* Must check return value! (MicroSD restart!) */
          if (MmcCommand(MMC_SET_BLOCKLEN|0x40, 512) != 0)
              goto tryagain;
          /* All OK return */
        #endif
    }

    // Init File System (FAT)

    /* Detects cards without partition table */
    memset(stream_buffer, 0, 512/2);
    if (InitFileSystem() == 0) {

	if ((USEX(SCI_AICTRL3) & CTRL3_NO_NUMFILES)) {
	    /* Startup is faster if we don't need to find number of files. */
	    USEX(SCI_AICTRL1) = 0x7fff;
	} else {
	    /* determine number of playable files */
	    USEX(SCI_AICTRL1) = OpenFile(0xffffU);
	    //putch(USEX(SCI_AICTRL1)>>8);
	    //putch(USEX(SCI_AICTRL1));
	}
	
      #if !defined(NO_RANDOM) /* if random play feature active */
	if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
            #ifdef SHUFFLE_PLAY /*shuffle 23 words*/
	  USEX(SCI_AICTRL0) = Shuffle(USEX(SCI_AICTRL1), USEX(SCI_AICTRL0));
            #else
                  /* random play (not shuffle!) */
	  USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL1));
            #endif
            }
          #endif
        
	/* Open the file to play. The file number is in AICTRL0. */
        if (OpenFile(USEX(SCI_AICTRL0) &= 0x7fffU
		     //max 32768 files in SCI-controlled version
            ) < 0) {
            #if defined(SKIP_ID3V2) /* 48 words */
              /* check for ID3v2 */
              /* An ID3v2 tag can be detected with the following pattern:
                 $49 44 33 yy yy xx zz zz zz zz */
              ReadDiskSector(fragments[0].start);
              if ((GetLong(0) & 0xffffffL) == 0x334449) {
                  /* skip in 16kB resolution, assume tags are smaller than 2MB */
                  /* 0.5kB resolution now */
                  register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
                  /* guard against small fragments */
                  if (skip >= fragments[0].size)
                      skip = fragments[0].size-1;
                  fragments[0].start += skip;
                  fragments[0].size  -= skip;
                  fatInfo.fileSize -= (u_int32)skip * 512;
              }
            #endif

              if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
                  CTRL3_PAUSE_BEFORE_PLAY) {
                  /* Mark file ready and go to pause state in UserHook */
                  USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
              }
              /*NEW: set FILE_READY always */
              USEX(SCI_AICTRL3) = CTRL3_FILE_READY | 
                  (USEX(SCI_AICTRL3) & ~CTRL3_AT_END);
            
            #if defined(PAUSE_AT_POWERON)
	      /*pause at power-on, then use pause after play*/
              if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
                  CTRL3_PAUSE_BEFORE_PLAY) {
                  USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY;
                  oldSpiState = spi_file;
                  spiState = spi_pause;
                  stream_save = stream_wr_pointer;
              } else {
                  spiState = spi_file;
              }
            #else /*PAUSE_AT_POWERON*/
		if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
		    CTRL3_PAUSE_BEFORE_PLAY) {
		    oldSpiState = spi_file;
		    spiState = spi_pause;
		    stream_save = stream_wr_pointer;
		} else {
		    spiState = spi_file;
		}
            #endif /*else PAUSE_AT_POWERON*/
            
            
            #ifdef HDATCHECK
              hdatCheck = 0;
            #endif
        } else {
        gofirst:
            USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
            MyReset();
        }
    } else {
        /* Initialize again */
        goto gofirst;//MyReset();
    }

    #if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
      /*otherwise keep it in powerdown until MMC found*/
      USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power */
    #endif
    
    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
    USEX(SCI_DECODE_TIME) = 0;
   
    parametric_x.resync = 32767; /* enable resyncs */
    stream_rd_pointer = stream_wr_pointer = (void*)stream_buffer;
    stream_rd_odd = 0;
    /* Main decode loop -- patched version of ReadFrame() does not fix
       known decoding problems, vs1053b-patches should also be integrated. */
    InitReadFrame();
    while (1) {
		register int j;
		if (
		#if defined(USE_FLAC)
		!(j = PatchFrame())
		#else
		!(j = ReadFrame())
		#endif
		) {
			play_frame();
		} else {
			InitReadFrame();
			needCheckNextHeader = 0;
		}
		UserHook();
    }
}
